from amuse.test.amusetest import TestWithMPI

from amuse.community.sse.interface import SSE
from amuse.community.mesa.interface import MESA
from amuse.community.evtwin.interface import EVtwin
from amuse.community.seba.interface import SeBa
from amuse.community.bse.interface import BSE

import numpy
import time

from amuse.units import units
from amuse import datamodel


class _TestStellarEvolutionCodes(TestWithMPI):

    def code_factory(self):
        self.skip("abstract test")

    def test1(self):
        instance = self.code_factory()

    def test2(self):
        for m in ([0.2, 1., 5., 25.] | units.MSun):
            instance = self.code_factory()()
            p = datamodel.Particle(mass=m)
            p2 = instance.particles.add_particle(p)
            self.assertAlmostEqual(p2.mass, p.mass)
            self.assertTrue(hasattr(p2, "radius"))
            self.assertTrue(hasattr(p2, "luminosity"))
            self.assertTrue(hasattr(p2, "age"))
            self.assertTrue(hasattr(p2, "stellar_type"))

    def test3(self):
        instance = self.code_factory()()
        p = datamodel.Particle(mass=1. | units.MSun)
        p2 = instance.particles.add_particle(p)
        self.assertAlmostEqual(p2.mass, p.mass)
        self.assertTrue(hasattr(p2, "radius"))
        self.assertTrue(hasattr(p2, "luminosity"))
        self.assertTrue(hasattr(p2, "age"))
        self.assertTrue(hasattr(p2, "stellar_type"))
        self.assertEqual(str(p2.stellar_type), "Main Sequence star")

    def test4(self):
        for m in ([0.2, 1., 5.] | units.MSun):
            instance = self.code_factory()()
            p = datamodel.Particle(mass=m)
            p2 = instance.particles.add_particle(p)
            self.assertAlmostEqual(p2.age, 0. | units.Myr)
            instance.evolve_model(1. | units.Myr)
            self.assertGreaterEqual(p2.age,  (1. | units.Myr))
            self.assertGreaterEqual(instance.model_time,  (1. | units.Myr))

    def test5(self):
        instance = self.code_factory()()
        p1 = datamodel.Particle(mass=1. | units.MSun)
        p2 = datamodel.Particle(mass=1. | units.MSun)
        p1 = instance.particles.add_particle(p1)
        instance.evolve_model(1. | units.Gyr)
        p2 = instance.particles.add_particle(p2)
        instance.evolve_model(2. | units.Gyr)
        self.assertGreaterEqual(p1.age, (1. | units.Gyr))
        self.assertGreaterEqual(p2.age, (1. | units.Gyr))
        self.assertLess(p2.age, (2. | units.Gyr))

    def test6(self):
        instance1 = self.code_factory()()
        p1 = datamodel.Particle(mass=1. | units.MSun)
        p2 = datamodel.Particle(mass=1. | units.MSun)
        p3 = datamodel.Particle(mass=1. | units.MSun)
        p1 = instance1.particles.add_particle(p1)
        instance1.evolve_model(10. | units.Myr)
        p2 = instance1.particles.add_particle(p2)
        instance1.evolve_model(20. | units.Myr)

        instance2 = self.code_factory()()
        p3 = instance2.particles.add_particle(p3)
        instance2.evolve_model(10. | units.Myr)

        self.assertAlmostEqual(p2.mass, p3.mass)
        self.assertAlmostEqual(p2.radius, p3.radius)
        self.assertAlmostEqual(p2.luminosity, p3.luminosity)

    def test7(self):
        instance1 = self.code_factory()()
        instance2 = self.code_factory()()
        p1 = datamodel.Particle(mass=1. | units.MSun)
        p1 = instance1.particles.add_particle(p1)
        p1.evolve_one_step()
        p2 = datamodel.Particle(mass=1. | units.MSun)
        p2 = instance2.particles.add_particle(p2)
        instance2.evolve_model(p1.age)
        self.assertAlmostEqual(p1.mass, p2.mass)
        self.assertAlmostEqual(p1.radius, p2.radius)
        self.assertAlmostEqual(p1.luminosity, p2.luminosity)

    def test8(self):
        instance1 = self.code_factory()()
        instance2 = self.code_factory()()
        p1 = datamodel.Particle(mass=1. | units.MSun)
        p1 = instance1.particles.add_particle(p1)
        p1.evolve_for(10. | units.Myr)
        p2 = datamodel.Particle(mass=1. | units.MSun)
        p2 = instance2.particles.add_particle(p2)
        instance2.evolve_model(10. | units.Myr)
        self.assertAlmostEqual(p1.mass, p2.mass)
        self.assertAlmostEqual(p1.radius, p2.radius)
        self.assertAlmostEqual(p1.luminosity, p2.luminosity)


"""
the following two tests will not work - this is to be fixed.

    def test6(self):
        instance=self.code_factory()()
        p1=datamodel.Particle(mass=1. | units.MSun)
        p2=datamodel.Particle(mass=1. | units.MSun,age=1.| units.Gyr)
        p1=instance.particles.add_particle(p1)
        p2=instance.particles.add_particle(p2)
        self.assertAlmostEqual(p1.age, (0. | units.Gyr) )
        self.assertAlmostEqual(p2.age, (1. | units.Gyr) )
        
    def test7(self):
        p1=datamodel.Particle(mass=8. | units.MSun)
        p2=datamodel.Particle(mass=8. | units.MSun,age=42. | units.Myr)

        instance1=self.code_factory()()
        p1=instance1.particles.add_particle(p1)
        instance2=self.code_factory()()
        p2=instance2.particles.add_particle(p2)
        
        instance1.evolve_model(44. | units.Myr)
        instance1.evolve_model(2. | units.Myr)

        self.assertalmostEqual( p1.mass,p2.mass)
"""


class TestSSECode(_TestStellarEvolutionCodes):
    def code_factory(self):
        return SSE

# class xTestBSECode(_TestStellarEvolutionCodes):
#    def code_factory(self):
#        return BSE


class TestMESACode(_TestStellarEvolutionCodes):
    def code_factory(self):
        try:
            MESA()
        except Exception as message:
            self.skip("Tried to instantiate a new object of the optional code with type '{0}', but this code is not available".format(MESA))
        return MESA


class TestSeBaCode(_TestStellarEvolutionCodes):
    def code_factory(self):
        return SeBa


class TestEVtwinCode(_TestStellarEvolutionCodes):
    def code_factory(self):
        return EVtwin



from amuse.test.amusetest import TestWithMPI, get_path_to_results
import sys
import os.path
from subprocess import PIPE, Popen
import numpy

from amuse.community.mesa_r15140.interface import MESA, MESAInterface

from amuse.support.exceptions import AmuseException
from amuse.units import units
from amuse.datamodel import Particles
from amuse.datamodel import Particle
from amuse.ext.spherical_model import EnclosedMassInterpolator


def set_mesa_paths_instance(instance):
    instance.set_MESA_paths(
        instance.default_path_to_inlist,
        instance.default_path_to_MESA,
        instance.default_path_to_MESA_data,
        instance.get_output_directory(),
        '',
        instance.default_tmp_dir
    )


class TestMESAInterface(TestWithMPI):

    def test1(self):
        print("Testing initialization of the interface...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        set_mesa_paths_instance(instance)
        instance.stop()

    def test2(self):
        print(
            "Testing get/set of metallicity (tests new ZAMS model "
            "implicitly)..."
        )
        print(
            "The first time this test will take quite some time"
            " to generate new starting models."
        )
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        set_mesa_paths_instance(instance)
        (metallicity, error) = instance.get_metallicity()
        self.assertEqual(0, error)
        self.assertEqual(0.02, metallicity)
        for x in [0.04, 0.02, 0.01, 0.00]:
            error = instance.set_metallicity(x)
            self.assertEqual(0, error)
            (metallicity, error) = instance.get_metallicity()
            self.assertEqual(0, error)
            self.assertEqual(x, metallicity)
        instance.stop()

    def test3(self):
        print("Testing basic operations: new_particle...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        (
            maximum_number_of_stars, error
        ) = instance.get_maximum_number_of_stars()
        self.assertEqual(0, error)
        self.assertEqual(maximum_number_of_stars, 10)
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        number_of_stars = 1
        for i in range(number_of_stars):
            (
                index_of_the_star, error
            ) = instance.new_particle(0.5+i*1.0/number_of_stars)
            self.assertEqual(0, error)
            self.assertEqual(index_of_the_star, i+1)
        instance.stop()

    def test4(self):
        print("Testing basic operations: evolve...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        (index_of_the_star, error) = instance.new_particle(1.0)
        self.assertEqual(0, error)
        self.assertEqual(index_of_the_star, 1)
        self.assertEqual(0, instance.commit_particles())

        initial_dt = 1.0e5
        dt_factor = 1.2

        instance.set_time_step(index_of_the_star, initial_dt)
        self.assertEqual([initial_dt, 0], list(instance.get_time_step(index_of_the_star).values()))

        self.assertEqual(0, instance.evolve_for(index_of_the_star, initial_dt))
        self.assertEqual(
            [initial_dt, 0],
            list(instance.get_age(index_of_the_star).values())
        )

        target_end_time = 3.0e5  # (years)
        self.assertEqual(
            0, instance.evolve_for(
                index_of_the_star, target_end_time-initial_dt
            )
        )

        self.assertTrue(
            instance.get_age(index_of_the_star)['age'] >= target_end_time
        )

        (L_of_the_star, error) = instance.get_luminosity(index_of_the_star)
        self.assertEqual(0, error)
        self.assertAlmostEqual(L_of_the_star, 0.725, 1)
        (M_of_the_star, error) = instance.get_mass(index_of_the_star)
        self.assertEqual(0, error)
        self.assertAlmostEqual(M_of_the_star, 1.000, 3)
        (T_of_the_star, error) = instance.get_temperature(index_of_the_star)
        self.assertEqual(0, error)
        self.assertAlmostEqual(T_of_the_star, 5650.998, -2)
        instance.stop()

    def slowtest5(self):
        print(
            "Testing evolve with varying Z (tests new ZAMS model "
            "implicitly)..."
        )
        print(
            "If the required starting models do not exist, this test will "
            "take quite some time to generate them."
        )
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        status = instance.initialize_code()
        self.assertEqual(status, 0)
        metallicities = [0.00, 0.01, 0.02, 0.04]
        luminosities = [1.717, 0.938, 0.725, 0.592]
        for (i, (Z, L)) in enumerate(zip(metallicities, luminosities)):
            status = instance.set_metallicity(Z)
            self.assertEqual(0, status)
            (index_of_the_star, status) = instance.new_pre_ms_particle(1.0)
            self.assertEqual(0, status)
            self.assertEqual(index_of_the_star, i+1)
            instance.evolve_for(index_of_the_star, 5.0e5)
            (
                L_of_the_star, status
            ) = instance.get_luminosity(index_of_the_star)
            self.assertEqual(0, status)
            self.assertAlmostEqual(L_of_the_star, L, 1)
        instance.stop()

    def test6(self):
        print("Testing MESA stop conditions...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        (value, error) = instance.get_max_age_stop_condition()
        self.assertEqual(0, error)
        self.assertEqual(1.0e36, value)
        for x in range(10, 14):
            error = instance.set_max_age_stop_condition(10 ** x)
            self.assertEqual(0, error)
            (value, error) = instance.get_max_age_stop_condition()
            self.assertEqual(0, error)
            self.assertEqual(10 ** x, value)
        (value, error) = instance.get_min_timestep_stop_condition()
        self.assertEqual(0, error)
        self.assertEqual(1.0e-6, value)
        for x in range(-9, -2):
            error = instance.set_min_timestep_stop_condition(10.0 ** x)
            self.assertEqual(0, error)
            (value, error) = instance.get_min_timestep_stop_condition()
            self.assertEqual(0, error)
            self.assertEqual(10.0 ** x, value)
        instance.stop()

    def test7(self):
        print("Testing MESA parameters...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        (index_of_the_star, error) = instance.new_particle(1.0)
        (value, error) = instance.get_mixing_length_ratio(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual(2.0, value)
        for x in [1.0, 1.5, 3.0]:
            error = instance.set_mixing_length_ratio(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_mixing_length_ratio(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)
        (value, error) = instance.get_semi_convection_efficiency(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        for x in [0.1, 0.04, 0.001]:
            error = instance.set_semi_convection_efficiency(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_semi_convection_efficiency(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)
        instance.stop()

    def test8(self):
        print("Testing MESA wind parameters...")
        instance = self.new_instance_of_an_optional_code(MESAInterface)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        (index_of_the_star, error) = instance.new_particle(1.0)
        (value, error) = instance.get_RGB_wind_scheme(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual('', value)
        for x, name in enumerate(['Reimers', 'Blocker', 'de jager', 'Dutch']):
            error = instance.set_RGB_wind_scheme(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_RGB_wind_scheme(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_AGB_wind_scheme(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual('', value)
        for x, name in enumerate(['Reimers', 'Blocker', 'de jager', 'Dutch']):
            error = instance.set_AGB_wind_scheme(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_AGB_wind_scheme(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_reimers_wind_efficiency(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_reimers_wind_efficiency(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_reimers_wind_efficiency(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_blocker_wind_efficiency(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_blocker_wind_efficiency(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_blocker_wind_efficiency(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_de_jager_wind_efficiency(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_de_jager_wind_efficiency(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_de_jager_wind_efficiency(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        (value, error) = instance.get_dutch_wind_efficiency(index_of_the_star)
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        for x in [0.0, 0.1, 0.5, 1.0]:
            error = instance.set_dutch_wind_efficiency(index_of_the_star, x)
            self.assertEqual(0, error)
            (value, error) = instance.get_dutch_wind_efficiency(index_of_the_star)
            self.assertEqual(0, error)
            self.assertEqual(x, value)

        instance.stop()


class TestMESA(TestWithMPI):
    def test1(self):
        print("Testing initialization and default MESA parameters...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        self.assertEqual(0.02 | units.no_unit, instance.parameters.metallicity)
        self.assertEqual(
            1.0e36 | units.julianyr, instance.parameters.max_age_stop_condition
        )
        instance.parameters.max_age_stop_condition = 1.0e2 | units.Myr
        self.assertEqual(
            1.0e2 | units.Myr, instance.parameters.max_age_stop_condition
        )
        instance.stop()

    def test2(self):
        # FIXME seems to be a mismatch between the definition of "yr" between
        # MESA and AMUSE, and time_step and age.
        print("Testing basic operations: evolve and get_...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        evo_star = instance.particles.add_particle(
            Particle(mass=1.0 | units.MSun)
        )
        self.assertEqual(len(instance.particles), 1)
        # instance.commit_particles()

        initial_dt = 1.0e5 | units.julianyr
        dt_factor = 1.2
        time_step = evo_star.time_step
        self.assertAlmostEqual(time_step, initial_dt)
        evo_star.evolve_one_step()
        age_of_the_star = evo_star.age
        self.assertAlmostEqual(age_of_the_star, initial_dt)

        target_end_time = 3.0e5 | units.julianyr
        evo_star.evolve_for(
            target_end_time - age_of_the_star
        )

        self.assertTrue(
            evo_star.age >= target_end_time
        )

        L_of_the_star = evo_star.luminosity
        self.assertAlmostEqual(L_of_the_star, 0.725 | units.LSun, 1)
        M_of_the_star = evo_star.mass
        self.assertAlmostEqual(M_of_the_star, 1.000 | units.MSun, 3)
        T_of_the_star = evo_star.temperature
        self.assertAlmostEqual(T_of_the_star, 5650.998 | units.K, -2)
        instance.stop()

    def test3(self):
        print("Testing basic operations: evolve_model and channels...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        stars = Particles(1)
        mass = 10. | units.MSun
        stars.mass = mass
        instance.particles.add_particles(stars)
        instance.commit_particles()

        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()
        self.assertEqual(stars[0].mass, mass)
        self.assertAlmostRelativeEqual(
            stars[0].luminosity, 5841. | units.LSun, 1
        )
        instance.evolve_model()
        from_code_to_model.copy()
        self.assertAlmostEqual(stars[0].mass, mass, 5)
        self.assertAlmostRelativeEqual(
            stars[0].luminosity, 5820.85 | units.LSun, 1
        )
        instance.stop()

    def slowtest4(self):
        print("Testing stellar type...")
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        stars = Particles(1)

        star = stars[0]
        star.mass = 5.0 | units.MSun

        instance.particles.add_particles(stars)
        instance.commit_particles()

        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()

        previous_type = 15 | units.stellar_type
        results = []
        current_time = 0 | units.Myr

        while current_time < (101 | units.Myr):
            if not star.stellar_type == previous_type:
                print((star.age, star.mass, star.stellar_type))
                results.append((star.age, star.mass, star.stellar_type))
                previous_type = star.stellar_type
            instance.evolve_model()
            from_code_to_model.copy()
            current_time = star.age

        self.assertEqual(len(results), 4)

        times = (
            0.0 | units.Myr,
            81.6 | units.Myr,
            99.9 | units.Myr,
            100.3 | units.Myr
        )
        for result, expected in zip(results, times):
            self.assertAlmostEqual(
                result[0].value_in(units.Myr), expected.value_in(units.Myr), 1
            )

        masses = (
            5.000 | units.MSun,
            5.000 | units.MSun,
            5.000 | units.MSun,
            5.000 | units.MSun
        )
        for result, expected in zip(results, masses):
            self.assertAlmostEqual(
                result[1].value_in(units.MSun),
                expected.value_in(units.MSun),
                3
            )

        types = (
            "Main Sequence star",
            "First Giant Branch",
            "Core Helium Burning",
            "First Asymptotic Giant Branch"
        )

        for result, expected in zip(results, types):
            self.assertEqual(str(result[2]), expected)

        instance.stop()

    def test5(self):
        print("Testing evolve_model for particle set...")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        masses = [0.5, 1.0] | units.MSun
        # FIXME updated max_age, as time step seems to be larger now
        max_age = 1.0e6 | units.julianyr
        number_of_stars = len(masses)
        stars = Particles(number_of_stars)
        stars.mass = masses
        instance.initialize_code()
        self.assertEqual(
            instance.parameters.max_age_stop_condition, 1e36 | units.julianyr
        )
        instance.parameters.max_age_stop_condition = max_age
        self.assertEqual(instance.parameters.max_age_stop_condition, max_age)
        instance.particles.add_particles(stars)
        instance.commit_particles()
        from_code_to_model = instance.particles.new_channel_to(stars)
        instance.evolve_model(end_time=0.5e6 | units.julianyr)
        from_code_to_model.copy()
        for i in range(number_of_stars):
            self.assertTrue(stars[i].age.value_in(units.julianyr) >= 0.5e6)
            self.assertTrue(stars[i].age <= max_age)
            self.assertTrue(stars[i].mass <= masses[i])
            self.assertTrue(stars[i].age+stars[i].time_step <= max_age)

        # FIXME this stopping condition doesn't seem implemented yet
        # self.assertRaises(
        #     AmuseException, instance.evolve_model, end_time=2*max_age,
        #     expected_message=(
        #         "Error when calling 'evolve_for' of a 'MESA', "
        #         "errorcode is -12, error is "
        #         "'Evolve terminated: Maximum age reached.'"
        #     )
        # )
        instance.stop()

    def test6(self):
        print("Test for obtaining the stellar structure model")
        stars = Particles(2)
        stars.mass = [1.0, 10.0] | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model()
        self.assertEqual(instance.particles.get_number_of_zones(), [3832, 2468])
        self.assertEqual(len(instance.particles[0].get_mass_profile()), 3832)
        self.assertAlmostEqual(
            instance.particles[0].get_mass_profile().sum(), 1.0
        )
        self.assertRaises(
            AmuseException, instance.particles.get_mass_profile,
            expected_message=(
                "Querying mass profiles of more than one particle at a time "
                "is not supported."
            )
        )
        print(instance.particles)
        self.assertEqual(
            len(instance.particles[1].get_density_profile()), 2468
        )
        self.assertIsOfOrder(
            instance.particles[0].get_radius_profile()[-1],
            1.0 | units.RSun)
        self.assertIsOfOrder(
            instance.particles[0].get_temperature_profile()[0],
            1.0e7 | units.K)
        self.assertIsOfOrder(
            instance.particles[0].get_luminosity_profile()[-1],
            1.0 | units.LSun)
        self.assertIsOfOrder(
            instance.particles[0].get_pressure_profile()[0],
            1.0e17 | units.barye)
        delta_mass = (
            instance.particles[0].get_mass_profile()
            * instance.particles[0].mass
        )
        radius1 = instance.particles[0].get_radius_profile()
        radius2 = radius1[:-1]
        radius2.prepend(0 | units.m)
        delta_radius_cubed = (radius1**3 - radius2**3)
        # FIXME the test below doesn't pass for [1]*3832
        # but it passes for [0.99888]*3832?
        # self.assertAlmostEqual(
        #     instance.particles[0].get_density_profile()
        #     / (delta_mass/(4./3.*numpy.pi*delta_radius_cubed)),
        #     [1]*3832, places=3
        # )
        self.assertAlmostEqual(
            instance.particles[1].get_mu_profile(), [0.62]*2468 | units.amu,
            places=1
        )
        instance.stop()

    def test7(self):
        print("Test for obtaining the stellar composition structure")
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model()
        number_of_zones = instance.particles.get_number_of_zones()[0]
        number_of_species = instance.particles.get_number_of_species()[0]
        composition = instance.particles[0].get_chemical_abundance_profiles()
        species_names = instance.particles[0].get_names_of_species()
        species_IDs = instance.particles[0].get_IDs_of_species()
        species_masses = instance.particles[0].get_masses_of_species()
        self.assertEqual(number_of_zones, 3832)
        self.assertEqual(number_of_species,  8)
        self.assertEqual(len(species_names), number_of_species)
        self.assertEqual(len(composition), number_of_species)
        self.assertEqual(len(composition[0]), number_of_zones)
        self.assertEqual(
            species_names,
            ['h1', 'he3', 'he4', 'c12', 'n14', 'o16', 'ne20', 'mg24']
        )
        self.assertEqual(
            species_IDs,
            # [2,    5,     6,     38,    51,    69,    114,    168]
            [1,    2,     3,      4,     5,     6,      7,      8]
        )
        self.assertAlmostEqual(
            species_masses,
            [
                1.0078250, 3.0160293, 4.0026032, 12.0, 14.0030740, 15.9949146,
                19.9924401, 23.9850417
            ] | units.amu,
            places=5
        )
        self.assertAlmostEqual(composition[:1, -1].sum(),  0.7)
        self.assertAlmostEqual(
            composition[1:3, -1].sum(),
            (0.3) - instance.parameters.metallicity
        )
        self.assertAlmostEqual(
            composition[3:, -1].sum(),
            instance.parameters.metallicity
        )
        self.assertAlmostEqual(
            composition.sum(axis=0),
            [1.0]*number_of_zones
        )
        instance.stop()

    def slowtest8(self):
        print(
            "Test for obtaining the stellar composition structure - "
            "evolved star with zero metalicity"
        )
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.parameters.metallicity = 0.0
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()
        instance.evolve_model(5.85 | units.Gyr)
        self.assertTrue(instance.particles[0].age >= 5.85 | units.Gyr)
        self.assertTrue(
            str(instance.particles[0].stellar_type) == "First Giant Branch"
        )
        number_of_zones = instance.particles.get_number_of_zones()[0]
        number_of_species = instance.particles.get_number_of_species()[0]
        composition = instance.particles[0].get_chemical_abundance_profiles()
        species_names = instance.particles[0].get_names_of_species()
        species_IDs = instance.particles[0].get_IDs_of_species()
        self.assertEqual(number_of_zones,    578)
        self.assertEqual(number_of_species,    8)
        self.assertEqual(len(species_names),  number_of_species)
        self.assertEqual(len(composition),    number_of_species)
        self.assertEqual(len(composition[0]), number_of_zones)
        self.assertEqual(
            species_names,
            ['h1', 'he3', 'he4', 'c12', 'n14', 'o16', 'ne20', 'mg24']
        )
        self.assertEqual(
            species_IDs,
            [2,    5,     6,     38,    51,    69,    114,    168]
        )
        self.assertAlmostEqual(
            composition[:1, number_of_zones-1].sum(),  0.76 | units.none
        )
        self.assertAlmostEqual(
            composition[1:3, number_of_zones-1].sum(),  0.24 | units.none
        )
        self.assertAlmostEqual(
            composition[3:, number_of_zones-1].sum(),  0.00 | units.none
        )
        self.assertAlmostEqual(
            composition.sum(axis=0), [1.0]*number_of_zones | units.none
        )
        self.assertAlmostEqual(
            composition[:1, 0].sum(),  0.00 | units.none
        )
        self.assertAlmostEqual(
            composition[1:3, 0].sum(),  1.00 | units.none
        )
        self.assertAlmostEqual(
            composition[3:, 0].sum(),  0.00 | units.none
        )
        instance.stop()

    def xtest9(self):
        print("Test for changing the stellar structure model")
        ##################################################################
        # Changing the density or radius in this manner is not supported #
        ##################################################################

        # star = Particles(1)
        # star.mass = 1.0 | units.MSun
        # instance = self.new_instance_of_an_optional_code(MESA)
        # if instance is None:
        #     print("MESA was not built. Skipping test.")
        #     return
        # instance.initialize_code()
        # instance.commit_parameters()
        # instance.particles.add_particles(star)
        # instance.commit_particles()
        # instance.evolve_model()

        # density_profile = instance.particles[0].get_density_profile()

        # self.assertRaises(
        #     AmuseException,
        #     instance.particles[0].set_density_profile,
        #     density_profile[2:],
        #     expected_message=(
        #         "The length of the supplied vector (3830) does not match the "
        #         "number of mesh zones of the star (3832)."
        #     )
        # )

        # mass_factor = 1.1
        # instance.particles[0].set_density_profile(mass_factor*density_profile)
        # self.assertAlmostRelativeEqual(
        #     instance.particles[0].get_density_profile(),
        #     density_profile*mass_factor, places=10
        # )
        # instance.particles.mass *= mass_factor
        # instance.evolve_model()

        # outer_radius = instance.particles[0].get_radius_profile()
        # inner_radius = outer_radius[:-1]
        # inner_radius.prepend(0 | units.m)
        # delta_radius_cubed = (outer_radius**3 - inner_radius**3)
        # integrated_mass = (
        #     4./3. * numpy.pi * delta_radius_cubed
        #     * instance.particles[0].get_density_profile()
        # ).sum()
        # self.assertAlmostRelativeEqual(
        #     integrated_mass, star.mass*mass_factor, places=3
        # )
        # instance.stop()

    def test10(self):
        print("Test for changing the stellar composition")
        star = Particles(1)
        star.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(star)
        instance.commit_particles()
        instance.evolve_model(0.3 | units.Myr)

        composition = instance.particles[0].get_chemical_abundance_profiles()
        k_surface = -1  # index to the outer mesh cell (surface)

        self.assertAlmostEqual(composition[:1, k_surface].sum(),  0.7)
        self.assertAlmostEqual(
            composition[1:3, k_surface].sum(),
            (0.3) - instance.parameters.metallicity)
        self.assertAlmostEqual(
            composition[3:, k_surface].sum(),
            instance.parameters.metallicity)

        he4_start = composition[2, k_surface]
        h1_start = composition[0, k_surface]

        h1_profile = composition[0] * 1
        he4_profile = composition[2] * 1

        instance.particles[0].set_control('okay_to_remesh', False)  # Turn of remeshing

        # Gradually and consistently increase helium and decrease hydrogen
        # abundances until reversed
        for alpha in numpy.arange(0.1, 1.1, 0.1):
            composition[0] = alpha * he4_profile + (1-alpha) * h1_profile
            composition[2] = (1-alpha) * he4_profile + alpha * h1_profile
            instance.particles[0].set_chemical_abundance_profiles(composition)
            instance.particles[0].evolve_for(1 | units.julianyr)
            composition = (
                instance.particles[0].get_chemical_abundance_profiles()
            )

        self.assertAlmostEqual(
            composition[2, k_surface], h1_start, 3)
        self.assertAlmostEqual(composition[0, k_surface], he4_start, 3)
        self.assertAlmostEqual(
            composition[3:, k_surface].sum(),
            0.020, 3)
        self.assertAlmostEqual(composition.sum(axis=0), 1.0)

        self.assertRaises(
            AmuseException,
            instance.particles[0].set_chemical_abundance_profiles,
            composition[:7],
            expected_message=(
                "The length of the supplied vector (7) does not match "
                "the number of chemical species of the star (8)."
            )
        )
        instance.stop()

    def test11(self):
        print(
            "Test evolve_model optional arguments: end_time and "
            "keep_synchronous"
        )
        stars = Particles(3)
        stars.mass = [1.0, 2.0, 3.0] | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(stars)
        instance.commit_particles()

        self.assertAlmostEqual(
            instance.particles.age, [0.0, 0.0, 0.0] | units.yr)

        print(
            "evolve_model without arguments: use shared timestep = "
            "0.99*min(particles.time_step)"
        )
        instance.evolve_model()
        self.assertAlmostEqual(
            instance.particles.age.in_(units.yr),
            [6350.98860534, 6350.98860534, 6350.98860534] | units.yr, 3
        )

        self.assertAlmostEqual(
            instance.model_time, 0.99 * 6415.0029 | units.julianyr, 3
        )

        print(
            "evolve_model with end_time: take timesteps, until end_time "
            "is reached exactly"
        )
        instance.evolve_model(15000 | units.yr)

        self.assertAlmostEqual(instance.model_time, 15000.0 | units.yr, 3)

        print(
            "evolve_model with keep_synchronous: use non-shared timestep, "
            "particle ages will typically diverge"
        )
        instance.evolve_model(keep_synchronous=False)

        self.assertAlmostEqual(
            instance.model_time, 15000.0 | units.yr, 3)
        instance.stop()

    def test12(self):
        print("Test for importing new stellar models")
        star = Particles(1)
        star.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(star)
        instance.commit_particles()
        instance.evolve_model(1 | units.Myr)

        number_of_zones = instance.particles[0].get_number_of_zones()
        composition = instance.particles[0].get_chemical_abundance_profiles(
            number_of_zones=number_of_zones)
        instance.new_particle_from_model(dict(
            mass=instance.particles[0].get_cumulative_mass_profile(
                number_of_zones=number_of_zones) * instance.particles[0].mass,
            radius=instance.particles[0].get_radius_profile(
                number_of_zones=number_of_zones),
            rho=instance.particles[0].get_density_profile(
                number_of_zones=number_of_zones),
            temperature=instance.particles[0].get_temperature_profile(
                number_of_zones=number_of_zones),
            luminosity=instance.particles[0].get_luminosity_profile(
                number_of_zones=number_of_zones),
            X_H=composition[0],
            X_He=composition[1] + composition[2],
            X_C=composition[3],
            X_N=composition[4],
            X_O=composition[5],
            X_Ne=composition[6],
            X_Mg=composition[7],
            X_Si=composition[7]*0.0,
            X_Fe=composition[7]*0.0), 10.0 | units.Myr)
        self.assertEqual(len(instance.particles), 2)
        self.assertEqual(len(instance.imported_stars), 1)

        self.assertIsOfOrder(
            instance.imported_stars[0].get_radius_profile()[-1],
            1.0 | units.RSun)
        self.assertIsOfOrder(
            instance.imported_stars[0].get_temperature_profile()[0],
            1.0e7 | units.K)
        self.assertIsOfOrder(
            instance.imported_stars[0].get_pressure_profile()[0],
            1.0e17 | units.barye)
        self.assertAlmostEqual(
            instance.imported_stars[0].get_mass_profile()[-1],
            instance.native_stars[0].get_mass_profile()[-1])

        print(instance.particles)
        instance.evolve_model(keep_synchronous=False)
        print(instance.particles)
        instance.stop()

    def slowtest13(self):
        print("Testing MESA wind parameters...")
        stars = Particles(9)
        stars.mass = 10.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()

        winds = {
            '': 0.0,
            'Reimers': 0.5,
            'Blocker': 0.1,
            'de Jager': 0.8,
            'Dutch': 0.8,
        }

        instance.commit_parameters()
        for i, (wind_scheme, scale) in enumerate(winds):
            instance.recommit_parameters()
            instance.particles.add_particle(stars[i])
            stars[i].RGB_wind_scheme = wind_scheme
            stars[i].set_RGB_wind_efficiency(scale)

            winds[wind_scheme] = stars[i].get_RGB_wind_efficiency() * 2

        for i, (wind_scheme, scale) in enumerate(winds):
            instance.recommit_parameters()
            instance.particles.add_particle(stars[i+4])
            stars[i+4].RGB_wind_scheme = wind_scheme
        instance.commit_particles()
        instance.evolve_model(keep_synchronous=False)
        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()
        self.assertAlmostEqual(stars[0].wind, 0.0 | units.MSun / units.yr)
        self.assertAlmostRelativeEqual(
            stars[1:5].wind,
            [
                4.59318475897e-10, 5.20742729636e-11, 1.05565558121e-09,
                3.62519254311e-09
            ] | units.MSun / units.yr, places=7)
        self.assertAlmostRelativeEqual(
            stars[5:].wind, 2.0 * stars[1:5].wind, places=7)
        instance.stop()

    def test14(self):
        print("Testing MESA wind parameters... (short version of slowtest13)")
        stars = Particles(3)
        stars.mass = 10.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        instance.initialize_code()
        instance.commit_parameters()
        for i, wind_efficiency in enumerate([0.0, 0.5, 1.0]):
            instance.particles.add_particle(stars[i])
            stars[i].RGB_wind_scheme = 'Reimers'
            stars[i].reimers_wind_efficiency = wind_efficiency
        instance.commit_particles()
        instance.evolve_model(keep_synchronous=False)

        from_code_to_model = instance.particles.new_channel_to(stars)
        from_code_to_model.copy()

        self.assertAlmostEqual(stars[0].mass_change, 0.0 | units.MSun / units.yr)
        self.assertAlmostRelativeEqual(
            stars[1].mass_change, 4.59318475897e-10 | units.MSun / units.yr, places=1)
        self.assertAlmostRelativeEqual(
            stars[2].mass_change, 2.0 * stars[1].mass_change, places=7)
        instance.stop()

    def xtest15(self):
        print("Testing MESA states")
        stars = Particles(2)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        print("First do everything manually:", end=' ')
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        # FIXME: ??
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.particles.add_particle(stars[0])
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()
        print("ok")

        print(
            "initialize_code(), commit_parameters(), (re)commit_particles(), "
            "and cleanup_code() should be called automatically:", end=' ')
        instance = MESA()
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        set_mesa_paths_instance(instance)
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.particles.add_particle(stars[0])
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.particles.add_particle(stars[1])
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')
        print("ok")

    def test16(self):
        print("Testing basic operations: evolve_one_step and evolve_for")
        stars = Particles(2)
        stars.mass = 1.0 | units.MSun
        instance = self.new_instance_of_an_optional_code(MESA)
        if instance is None:
            print("MESA was not built. Skipping test.")
            return
        set_mesa_paths_instance(instance)
        se_stars = instance.particles.add_particles(stars)

        for i in range(3):
            se_stars[0].evolve_one_step()
        self.assertAlmostEqual(se_stars.age, [364000.0, 0] | units.julianyr)

        se_stars[1].evolve_for(se_stars[0].age)
        self.assertAlmostRelativeEqual(se_stars[0].age, se_stars[1].age)
        self.assertAlmostRelativeEqual(
            se_stars[0].luminosity, se_stars[1].luminosity, 2)
        self.assertAlmostRelativeEqual(
            se_stars[0].radius, se_stars[1].radius, 2)
        self.assertAlmostRelativeEqual(
            se_stars[0].temperature, se_stars[1].temperature, 2)
        instance.stop()

    def test18(self):
        print("Testing MESA mass_change (User-specified wind/accretion)")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)

        star = instance.particles.add_particle(Particle(mass=1 | units.MSun))
        star.mass_change = 1.0e-8 | units.MSun / units.julianyr  # positive -> accretion
        star.evolve_for(1.0e5 | units.julianyr)

        self.assertAlmostRelativeEqual(
            star.mass_change, 1.0e-8 | units.MSun / units.julianyr)
        self.assertAlmostRelativeEqual(
            star.wind, 1.0e-8 | units.MSun / units.julianyr, 3)
        self.assertAlmostRelativeEqual(star.age, 1.0e5 | units.julianyr)
        self.assertAlmostRelativeEqual(star.mass, 1.0010 | units.MSun)

        star.mass_change = -1.0e-8 | units.MSun / units.julianyr  # negative -> wind
        star.evolve_for(1 | units.s)

        self.assertAlmostRelativeEqual(
            star.mass_change, -1.0e-8 | units.MSun / units.julianyr)
        self.assertAlmostRelativeEqual(
            star.wind, -1.0e-8 | units.MSun / units.julianyr, 3)

        star.evolve_for(1.0e5 | units.julianyr)
        self.assertAlmostRelativeEqual(star.age, 2.e5 | units.julianyr, 3)
        self.assertAlmostRelativeEqual(star.mass,  1.00 | units.MSun, 3)
        print(star.as_set())
        instance.stop()

    def slowtest19a(self):
        print("Testing MESA core mass")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        star = instance.particles.add_particle(
            Particle(mass=3 | units.MSun))
        star.evolve_for(330 | units.Myr)
        for i in range(10):
            star.evolve_for(10 | units.Myr)
            index = numpy.searchsorted(
                star.get_chemical_abundance_profiles(number_of_species=1)[0],
                1.0e-4
            )
            h_poor_mass = EnclosedMassInterpolator(
                radii=star.get_radius_profile(),
                densities=star.get_density_profile()
            ).enclosed_mass[index].as_quantity_in(units.MSun)
            print(h_poor_mass, star.core_mass)
            self.assertAlmostEqual(star.core_mass, h_poor_mass, 2)
        instance.stop()

    def test19b(self):
        print("Testing MESA core mass (short version of slowtest19a)")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        star = instance.particles.add_particle(Particle(mass=3 | units.MSun))
        star.evolve_one_step()
        index = numpy.searchsorted(
            star.get_chemical_abundance_profiles(
                number_of_species=1)[0], 1.0e-4
        )
        h_poor_mass = EnclosedMassInterpolator(
            radii=star.get_radius_profile(),
            densities=star.get_density_profile()
        ).enclosed_mass[index].as_quantity_in(units.MSun)
        self.assertAlmostEqual(star.core_mass, h_poor_mass, 2)
        instance.stop()

    def test20(self):
        # FIXME seems to fail on yr/julianyr error
        print("Testing MESA pre-main-sequence star")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        star = instance.pre_ms_stars.add_particle(
            Particle(mass=1.0 | units.MSun))

        self.assertAlmostEqual(
            star.time_step.in_(units.julianyr), 1.0e-5 | units.julianyr)
        star.evolve_one_step()
        self.assertAlmostEqual(
            star.age.in_(units.julianyr), 1.0e-5 | units.julianyr)

        instance.evolve_model(1.0 | units.julianyr)
        self.assertEqual(star.stellar_type, 17 | units.stellar_type)
        self.assertEqual(str(star.stellar_type), "Pre-main-sequence Star")
        self.assertTrue(star.age >= 1.0 | units.julianyr)
        # 4408.57819467 K
        self.assertTrue(star.temperature < 4500 | units.K)
        # 1.33431386707 LSun
        # FIXME self.assertTrue(star.luminosity > 10 | units.LSun)
        # 1.98009060937 RSun
        # FIXME self.assertTrue(star.radius > 2 | units.RSun)
        instance.stop()

    def slowtest21(self):
        print("Testing MESA calculate_core_mass")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        star = instance.particles.add_particle(Particle(mass=40 | units.MSun))
        instance.evolve_model(4.56 | units.Myr)
        total_core_mass = star.calculate_core_mass()
        self.assertAlmostRelativeEqual(star.core_mass, total_core_mass, 2)
        self.assertTrue(
            star.calculate_core_mass(core_H_abundance_limit=1.0e-2)
            > total_core_mass
        )
        self.assertEqual(
            star.calculate_core_mass(core_H_abundance_limit=1.0e-4),
            total_core_mass)
        self.assertTrue(
            star.calculate_core_mass(core_H_abundance_limit=1.0e-6)
            < total_core_mass)
        self.assertAlmostRelativeEqual(
            star.calculate_core_mass(core_H_abundance_limit=0.8), star.mass, 2)

        h1_core_mass = star.calculate_core_mass(species=["h1"])
        he3_core_mass = star.calculate_core_mass(species=["he3"])
        he4_core_mass = star.calculate_core_mass(species=["he4"])
        c12_core_mass = star.calculate_core_mass(species=["c12"])
        n14_core_mass = star.calculate_core_mass(species=["n14"])
        o16_core_mass = star.calculate_core_mass(species=["o16"])
        ne20_core_mass = star.calculate_core_mass(species=["ne20"])
        mg24_core_mass = star.calculate_core_mass(species=["mg24"])
        metal_core_mass = star.calculate_core_mass(
            species=["c12", "n14", "o16", "ne20", "mg24"])
        print(
            h1_core_mass, he3_core_mass, he4_core_mass, c12_core_mass,
            n14_core_mass, o16_core_mass
        )
        print(ne20_core_mass, mg24_core_mass, metal_core_mass)
        self.assertAlmostRelativeEqual(star.core_mass, total_core_mass, 2)
        instance.stop()
        self.assertAlmostRelativeEqual(total_core_mass, he4_core_mass, 1)
        self.assertAlmostRelativeEqual(
            total_core_mass, he4_core_mass + metal_core_mass, 4)
        self.assertAlmostRelativeEqual(
            total_core_mass, he4_core_mass + metal_core_mass + h1_core_mass, 7)
        self.assertAlmostRelativeEqual(
            metal_core_mass,
            (
                c12_core_mass + n14_core_mass + o16_core_mass + ne20_core_mass
                + mg24_core_mass
            ), 7)
        self.assertAlmostEqual(he3_core_mass, 0 | units.MSun)

    def test22(self):
        print("Testing MESA calculate_core_mass (short version of slowtest21)")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        star = instance.particles.add_particle(Particle(mass=1 | units.MSun))
        instance.evolve_model(0.2 | units.Gyr)  # VERY short, for test speed up
        central_hydrogen_abundance = star.get_chemical_abundance_profiles(
        )[0][0]
        print(central_hydrogen_abundance)
        self.assertTrue(
            central_hydrogen_abundance < 0.69)  # some hydrogen is burned
        self.assertTrue(
            central_hydrogen_abundance > 0.68)  # ... but not that much yet
        self.assertEqual(
            star.calculate_core_mass(core_H_abundance_limit=0.67),
            0 | units.MSun)
        # FIXME the test below should give a "core mass" nearly equal to the
        # stellar mass, but it is a bit too small, maybe missing one layer/zone?
        # self.assertAlmostEqual(
        #     star.calculate_core_mass(core_H_abundance_limit=0.7),
        #     1 | units.MSun,
        #     3
        # )

        # For test speed up, we use a weird core_H_abundance_limit to define
        # the "hydrogen exhausted core"
        # FIXME updated values in the rest of the test
        limit = 0.69
        expected_core_mass = 0.0288211177139 | units.MSun
        self.assertAlmostEqual(
            star.calculate_core_mass(core_H_abundance_limit=limit),
            expected_core_mass, 3)

        h1_core_mass = star.calculate_core_mass(
            species=["h1"], core_H_abundance_limit=limit)
        he3_core_mass = star.calculate_core_mass(
            species=["he3"], core_H_abundance_limit=limit)
        he4_core_mass = star.calculate_core_mass(
            species=["he4"], core_H_abundance_limit=limit)
        c12_core_mass = star.calculate_core_mass(
            species=["c12"], core_H_abundance_limit=limit)
        n14_core_mass = star.calculate_core_mass(
            species=["n14"], core_H_abundance_limit=limit)
        o16_core_mass = star.calculate_core_mass(
            species=["o16"], core_H_abundance_limit=limit)
        ne20_core_mass = star.calculate_core_mass(
            species=["ne20"], core_H_abundance_limit=limit)
        mg24_core_mass = star.calculate_core_mass(
            species=["mg24"], core_H_abundance_limit=limit)
        metal_core_mass = star.calculate_core_mass(
            species=["c12", "n14", "o16", "ne20", "mg24"],
            core_H_abundance_limit=limit)
        instance.stop()
        self.assertAlmostRelativeEqual(
            h1_core_mass, expected_core_mass*0.685, 2)
        self.assertAlmostRelativeEqual(
            he4_core_mass, expected_core_mass*0.29, 2)
        self.assertAlmostRelativeEqual(
            metal_core_mass, expected_core_mass*0.02, 1)
        self.assertAlmostRelativeEqual(
            expected_core_mass,
            he4_core_mass + he3_core_mass + metal_core_mass + h1_core_mass, 7)
        self.assertAlmostRelativeEqual(
            metal_core_mass,
            (
                c12_core_mass + n14_core_mass + o16_core_mass + ne20_core_mass
                + mg24_core_mass
            ), 7)
        self.assertAlmostEqual(he3_core_mass, 5.91e-6 | units.MSun, 5)

    def test23(self):
        print("Testing MESA central_temperature and central_density")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        stars = instance.particles.add_particles(
            Particles(mass=[0.1, 1, 10] | units.MSun))
        self.assertIsOfOrder(
            stars.central_temperature, [4e6, 13e6, 31e6] | units.K)
        self.assertIsOfOrder(
            stars.central_density, [400, 77, 9] | units.g * units.cm**-3)
        instance.stop()

    def test24(self):
        print("Testing MESA calculate_helium_exhausted_core_mass")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        star = instance.particles.add_particle(Particle(mass=2 | units.MSun))

        composition = star.get_chemical_abundance_profiles()
        # Mimic hydrogen exhausted core:
        composition[2, :100] = composition[2, :100] + composition[0, :100]
        composition[0, :100] = 0
        # Mimic helium exhausted core:
        carbon_oxygen = composition[:6, :50].sum(axis=0)
        composition[3, :50] = carbon_oxygen * 0.6
        composition[5, :50] = carbon_oxygen * 0.4
        composition[1:3, :50] = 0
        composition[4, :50] = 0
        star.set_chemical_abundance_profiles(composition)

        self.assertAlmostRelativeEqual(
            star.calculate_core_mass(),
            star.mass * star.get_cumulative_mass_profile()[100], 1)
        self.assertAlmostRelativeEqual(
            star.calculate_helium_exhausted_core_mass(),
            star.mass * star.get_cumulative_mass_profile()[50], 1)

        core_mass = star.calculate_helium_exhausted_core_mass(
            split_species=False)
        core_mass_by_species = star.calculate_helium_exhausted_core_mass(
            split_species=True)
        carbon_mass_in_core, oxygen_mass_in_core = \
            star.calculate_helium_exhausted_core_mass(
                split_species=True, species=["c12", "o16"])

        self.assertEqual(
            len(core_mass_by_species), len(star.get_names_of_species())
        )
        instance.stop()
        self.assertAlmostRelativeEqual(core_mass, core_mass_by_species.sum())
        self.assertEqual(core_mass_by_species[0:3].sum(), 0 | units.MSun)
        self.assertEqual(core_mass_by_species[4], 0 | units.MSun)
        self.assertEqual(core_mass_by_species[3], carbon_mass_in_core)
        self.assertEqual(core_mass_by_species[5], oxygen_mass_in_core)

    def test25(self):
        print("Testing MESA accretion")
        instance = self.new_instance_of_an_optional_code(MESA)
        set_mesa_paths_instance(instance)
        star = instance.particles.add_particle(Particle(mass=2 | units.MSun))

        self.assertEqual(star.get_accrete_same_as_surface(), True)
        star.set_accrete_same_as_surface(False)
        self.assertEqual(star.get_accrete_same_as_surface(), False)

        self.assertEqual(
            star.get_accrete_composition_non_metals(), {'h1': 0.0, 'h2': 0.0, 'he3': 0.0, 'he4': 0.0}
        )
        self.assertEqual(star.get_accrete_composition_metals_identifier(), -1)
        self.assertEqual(star.get_accrete_composition_metals(), {'li': 0.0,
                                                                'be': 0.0,
                                                                'b': 0.0,
                                                                'c': 0.0,
                                                                'n': 0.0,
                                                                'o': 0.0,
                                                                'f': 0.0,
                                                                'ne': 0.0,
                                                                'na': 0.0,
                                                                'mg': 0.0,
                                                                'al': 0.0,
                                                                'si': 0.0,
                                                                'p': 0.0,
                                                                's': 0.0,
                                                                'cl': 0.0,
                                                                'ar': 0.0,
                                                                'k': 0.0,
                                                                'ca': 0.0,
                                                                'sc': 0.0,
                                                                'ti': 0.0,
                                                                'v': 0.0,
                                                                'cr': 0.0,
                                                                'mn': 0.0,
                                                                'fe': 0.0,
                                                                'co': 0.0,
                                                                'ni': 0.0,
                                                                'cu': 0.0,
                                                                'zn': 0.0})
        print("Accreting 75% deuterium", end=' ')
        print("and 25% iron")
        star.set_accrete_composition_non_metals(h2=0.75)
        star.set_accrete_composition_metals_identifier(0)  # i.e. specified below:
        star.set_accrete_composition_metals(fe=1.0)  # These must add to 1.0
        self.assertEqual(
            star.get_accrete_composition_non_metals()['h2'], 0.75)
        self.assertEqual(star.get_accrete_composition_metals_identifier(), 0)
        self.assertEqual(
            star.get_accrete_composition_metals()['fe'], 1.0)  # Metals must now sum to 1.0
                                                              # Z is set by the 1 - sum(non_metals)

        star.mass_change = 1.0e-8 | units.MSun / units.yr
        star.time_step = 0.1 | units.yr
        instance.evolve_model(1 | units.yr)
        composition = star.get_chemical_abundance_profiles()
        species = star.get_names_of_species()
        print("Both deuterium and iron are not in the current net,")
        print(
            f"so have been added to {species[0]} and {species[-1]}"
        )
        self.assertAlmostEqual(composition[:, -1], [0.75, 0, 0, 0, 0, 0, 0, 0.25])
        instance.stop()



from amuse.community import *
from amuse.test.amusetest import TestWithMPI

from amuse.community.rebound.interface import ReboundInterface
from amuse.community.rebound.interface import Rebound
import math
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class ReboundInterfaceTests(TestWithMPI):

    def test1(self):
        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        instance.initialize_code()

        res1, error = instance.new_particle(mass=11.0, radius=2.0, x=1.0, y=2.0, z=3.0, vx=4.0, vy=5.0, vz=6.0)
        res2, error = instance.new_particle(mass=21.0, radius=5.0, x=10.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        self.assertEqual(error, 0)
        self.assertEqual(res1, 0)
        self.assertEqual(res2, 1)

        m1, error = instance.get_mass(res1)
        self.assertEqual(error, 0)
        self.assertEqual(m1, 11)

        m2, error = instance.get_mass(res2)
        self.assertEqual(error, 0)
        self.assertEqual(m2, 21)

        x, y, z, error = instance.get_position(res1)
        self.assertEqual(error, 0)
        self.assertEqual(x, 1)
        self.assertEqual(y, 2)
        self.assertEqual(z, 3)

        vx, vy, vz, error = instance.get_velocity(res1)
        self.assertEqual(error, 0)
        self.assertEqual(vx, 4)
        self.assertEqual(vy, 5)
        self.assertEqual(vz, 6)
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        instance.initialize_code()

        instance.new_particle([10, 20], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1])
        retrieved_state = instance.get_state(0)

        self.assertEqual(10.0,  retrieved_state['mass'])

        retrieved_state = instance.get_state([0, 1])
        self.assertEqual(10.0,  retrieved_state['mass'][0])
        self.assertEqual(20.0,  retrieved_state['mass'][1])
        instance.cleanup_code()
        instance.stop()

    def test3(self):

        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        instance.initialize_code()

        indices, error = instance.new_particle([10, 20, 30], [1, 2, 3], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [1, 0, 1])
        n, error = instance.get_number_of_particles()
        self.assertEqual(n, 3)
        error = instance.delete_particle(indices[1])
        n, error = instance.get_number_of_particles()
        self.assertEqual(n, 2)
        instance.commit_particles()

        retrieved_state = instance.get_state([0, 2])
        self.assertEqual(10.0,  retrieved_state['mass'][0])
        self.assertEqual(30.0,  retrieved_state['mass'][1])
        mass, error = instance.get_mass(1)
        self.assertEqual(error, -1)
        instance.cleanup_code()
        instance.stop()

    def test4(self):

        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        instance.initialize_code()
        error = 0
        integrator = instance.get_integrator()
        self.assertEqual(error, 0)
        self.assertEqual("ias15", integrator)
        error = instance.set_integrator("whfast")
        self.assertEqual(error, 0)
        integrator = instance.get_integrator()
        self.assertEqual(error, 0)
        self.assertEqual("whfast", integrator)
        instance.cleanup_code()
        instance.stop()

    def test5(self):
        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        # Set up an equal-mass binary on a circular orbit:
        self.assertEqual([0, 0], list(instance.new_particle(0.5,  0.5, 0, 0,  0, 0.5, 0, 0.01).values()))
        self.assertEqual([1, 0], list(instance.new_particle(0.5,  -0.5, 0, 0,  0, -0.5, 0, 0.01).values()))
        self.assertEqual(0, instance.commit_particles())

        self.assertEqual(0, instance.evolve_model(math.pi))
        for result, expected in zip(list(instance.get_position(0).values()), [-0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 3)
        for result, expected in zip(list(instance.get_position(1).values()), [0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 3)

        self.assertEqual(0, instance.evolve_model(2 * math.pi))
        for result, expected in zip(list(instance.get_position(0).values()), [0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 3)
        for result, expected in zip(list(instance.get_position(1).values()), [-0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 3)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test6(self):
        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        index, err = instance.new_particle(0.5,  0.5, 0, 0,  0, 0.5, 0, 0.01, 1)
        self.assertEqual(-10, err)

        instance.cleanup_code()
        instance.stop()

    def test7(self):

        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        instance.initialize_code()

        instance.set_eps2(0.1 * 0.1)

        eps2 = instance.get_eps2()['epsilon_squared']

        self.assertEqual(0.1 * 0.1, eps2)
        instance.cleanup_code()
        instance.stop()

    def test8(self):

        instance = self.new_instance_of_an_optional_code(ReboundInterface)
        instance.initialize_code()

        instance.set_boundary("none")
        boundary_type = instance.get_boundary()
        self.assertEqual("none", boundary_type)

        instance.set_boundary("periodic")
        boundary_type = instance.get_boundary()
        self.assertEqual("periodic", boundary_type)

        instance.set_boundary_size(42.1)
        boundary_size = instance.get_boundary_size()['boundary_size']
        self.assertEqual(42.1, boundary_size)

        instance.cleanup_code()
        instance.stop()


class TestRebound(TestWithMPI):
    def new_system_of_sun_and_earth(self):
        stars = datamodel.Stars(2)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        return stars

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        interface = self.new_instance_of_an_optional_code(Rebound, convert_nbody)
        interface.initialize_code()
        interface.parameters.epsilon_squared = 0.0 | units.AU**2
        interface.dt_dia = 5000

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]

        interface.particles.add_particles(stars)

        interface.evolve_model(365.0 | units.day)
        interface.particles.copy_values_of_all_attributes_to(stars)

        position_at_start = earth.position.value_in(units.AU)[0]
        position_after_full_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostRelativeEqual(position_at_start, position_after_full_rotation, 6)

        interface.evolve_model(365.0 + (365.0 / 2) | units.day)

        self.assertAlmostRelativeEqual(interface.model_time, 365.0 + (365.0 / 2) | units.day, 3)
        interface.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostRelativeEqual(-position_at_start, position_after_half_a_rotation, 3)

        interface.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)
        self.assertAlmostRelativeEqual(interface.model_time, 365.0 + (365.0 / 2) + (365.0 / 4) | units.day, 3)

        interface.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[1]
        self.assertAlmostRelativeEqual(-position_at_start, position_after_half_a_rotation, 3)

        interface.cleanup_code()

        interface.stop()

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = self.new_instance_of_an_optional_code(Rebound, convert_nbody)
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.0 | units.AU**2
        instance.dt_dia = 5000

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]
        instance.particles.add_particles(stars)

        for x in range(1, 500, 10):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            stars.savepoint()

        if HAS_MATPLOTLIB:
            figure = pyplot.figure()
            plot = figure.add_subplot(1, 1, 1)

            x_points = earth.get_timeline_of_attribute("x")
            y_points = earth.get_timeline_of_attribute("y")

            x_points_in_AU = [t_x[1].value_in(units.AU) for t_x in x_points]
            y_points_in_AU = [t_x1[1].value_in(units.AU) for t_x1 in y_points]

            plot.scatter(x_points_in_AU, y_points_in_AU, color="b", marker='o')

            plot.set_xlim(-1.5, 1.5)
            plot.set_ylim(-1.5, 1.5)

            test_results_path = self.get_path_to_results()
            output_file = os.path.join(test_results_path, "rebound-earth-sun2.svg")
            figure.savefig(output_file)

        instance.cleanup_code()
        instance.stop()

    def test3(self):
        particles = datamodel.Particles(7)
        particles.mass = 0.001 | nbody_system.mass
        particles.radius = 0.1 | nbody_system.length
        particles.x = [-101.0, -100.0, -0.5, 0.5, 100.0, 101.0, 104.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [[2, 0, 0], [-2, 0, 0]]*3 + [[-4, 0, 0]] | nbody_system.speed

        instance = self.new_instance_of_an_optional_code(Rebound)
        instance.initialize_code()
        instance.parameters.set_defaults()
        instance.particles.add_particles(particles)
        collisions = instance.stopping_conditions.collision_detection
        collisions.enable()
        instance.evolve_model(1.0 | nbody_system.time)

        print(instance.particles)
        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 0.5 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 3)
        self.assertEqual(len(collisions.particles(1)), 3)
        self.assertEqual(len(particles - collisions.particles(0) - collisions.particles(1)), 1)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <=
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True, True, True])

        sticky_merged = datamodel.Particles(len(collisions.particles(0)))
        sticky_merged.mass = collisions.particles(0).mass + collisions.particles(1).mass
        sticky_merged.radius = collisions.particles(0).radius
        for p1, p2, merged in zip(collisions.particles(0), collisions.particles(1), sticky_merged):
            merged.position = (p1 + p2).center_of_mass()
            merged.velocity = (p1 + p2).center_of_mass_velocity()

        print(instance.model_time)
        print(instance.particles)
        instance.particles.remove_particles(collisions.particles(0) + collisions.particles(1))
        instance.particles.add_particles(sticky_merged)
        print(instance.particles)
        instance.evolve_model(1.0 | nbody_system.time)
        print()
        print(instance.model_time)
        print(instance.particles)
        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 1.0 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 1)
        self.assertEqual(len(collisions.particles(1)), 1)
        self.assertEqual(len(instance.particles - collisions.particles(0) - collisions.particles(1)), 2)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <=
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True])
        instance.stop()

    def test4(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        instance = self.new_instance_of_an_optional_code(Rebound, convert_nbody)
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.0 | units.AU**2
        instance.dt_dia = 5000

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]
        instance.particles.add_particles(stars)
        self.assertAlmostRelativeEquals(instance.kinetic_energy, stars.kinetic_energy())
        self.assertAlmostRelativeEquals(instance.potential_energy, stars.potential_energy(), 10)

        instance.stop()

    def test5(self):
        instance = self.new_instance_of_an_optional_code(Rebound)
        instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2

        particles = datamodel.Particles(2)
        particles.position = ([0, 0, 0], [1, 0, 0]) | nbody_system.length
        particles.velocity = ([-1, 0, 0], [2, 0, 0]) | nbody_system.speed
        particles.radius = 0 | nbody_system.length
        particles.mass = 0.1 | nbody_system.mass

        instance.particles.add_particles(particles)
        instance.stopping_conditions.out_of_box_detection.enable()
        instance.parameters.stopping_conditions_out_of_box_size = 2 | nbody_system.length
        instance.parameters.stopping_conditions_out_of_box_use_center_of_mass = False
        instance.evolve_model(1 | nbody_system.time)
        self.assertTrue(instance.stopping_conditions.out_of_box_detection.is_set())
        self.assertEqual(len(instance.stopping_conditions.out_of_box_detection.particles(0)), 1)
        self.assertEqual(instance.stopping_conditions.out_of_box_detection.particles(0)[0].key, particles[1].key)
        instance.stop()

    def test6(self):
        instance = self.new_instance_of_an_optional_code(Rebound)
        instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2

        particles = datamodel.Particles(2)
        particles.position = ([0, 0, 0], [1, 0, 0]) | nbody_system.length
        particles.velocity = ([-1, 0, 0], [2, 0, 0]) | nbody_system.speed
        particles.radius = 0 | nbody_system.length
        particles.mass = 0.1 | nbody_system.mass

        instance.particles.add_particles(particles)
        self.assertAlmostRelativeEquals(instance.kinetic_energy, particles.kinetic_energy())
        self.assertAlmostRelativeEquals(instance.potential_energy, particles.potential_energy(G=nbody_system.G))
        instance.stop()

    def test7(self):
        instance = self.new_instance_of_an_optional_code(Rebound)
        instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2
        subset0 = 0
        subset1 = instance.new_subset()
        print(subset1)
        self.assertEqual(subset1, 1)

        particles = datamodel.Particles(2)
        particles.position = ([0, 0, 0], [1, 0, 0]) | nbody_system.length
        particles.velocity = ([-1, 0, 0], [2, 0, 0]) | nbody_system.speed
        particles.radius = 0 | nbody_system.length
        particles.mass = 0.1 | nbody_system.mass

        particles2 = datamodel.Particles(2)
        particles2.position = ([0, 0, 0], [1, 0, 0]) | nbody_system.length
        particles2.velocity = ([-1, 0, 0], [2, 0, 0]) | nbody_system.speed
        particles2.radius = 0 | nbody_system.length
        particles2.mass = 2 | nbody_system.mass
        particles2.subset = subset1

        instance.particles.add_particles(particles)
        instance.particles.add_particles(particles2)
        print(instance.particles)
        self.assertAlmostRelativeEquals(instance.kinetic_energy, particles.kinetic_energy())
        self.assertAlmostRelativeEquals(instance.potential_energy, particles.potential_energy(G=nbody_system.G))
        self.assertAlmostRelativeEquals(instance.get_kinetic_energy(subset1), particles2.kinetic_energy())
        self.assertAlmostRelativeEquals(instance.get_potential_energy(subset1), particles2.potential_energy(G=nbody_system.G))
        # instance.stop()

    def test8(self):
        instance = self.new_instance_of_an_optional_code(Rebound)
        instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2
        subset0 = 0
        subset1 = instance.new_subset()
        print(subset1)
        self.assertEqual(subset1, 1)

        particles = datamodel.Particles(2)
        particles.position = ([0, 0, 0], [1, 0, 0]) | nbody_system.length
        particles.velocity = ([0, 0, 0], [0, 0.5, 0]) | nbody_system.speed
        particles.radius = 0 | nbody_system.length
        particles.mass = 0.1 | nbody_system.mass

        particles2 = datamodel.Particles(2)
        particles2.position = ([0, 0, 0], [2, 0, 0]) | nbody_system.length
        particles2.velocity = ([0, 0, 0], [0, 1, 0]) | nbody_system.speed
        particles2.radius = 0 | nbody_system.length
        particles2.mass = 1 | nbody_system.mass
        particles2.subset = subset1

        instance.particles.add_particles(particles)
        instance.particles.add_particles(particles2)
        instance.evolve_model(0.1 | nbody_system.time)
        self.assertAlmostRelativeEquals(instance.kinetic_energy, particles.kinetic_energy(), 2)
        self.assertAlmostRelativeEquals(instance.potential_energy, particles.potential_energy(G=nbody_system.G), 2)
        self.assertAlmostRelativeEquals(instance.get_kinetic_energy(subset1), particles2.kinetic_energy(), 2)
        self.assertAlmostRelativeEquals(instance.get_potential_energy(subset1), particles2.potential_energy(G=nbody_system.G), 2)
        particles_evolved = instance.particles.copy()
        instance.stop()

        instance1 = self.new_instance_of_an_optional_code(Rebound)
        instance1.parameters.epsilon_squared = 0.0 | nbody_system.length**2
        particles1c = particles.copy()
        instance1.particles.add_particles(particles1c)
        instance1.evolve_model(0.1 | nbody_system.time)
        particles1evolved = particles_evolved[particles_evolved.subset == 0]
        print("HHH:", particles_evolved.subset)
        print("p2e:", particles1evolved)
        print(instance1.particles)
        self.assertAlmostRelativeEquals(instance1.particles.position, particles1evolved.position, 10)
        self.assertAlmostRelativeEquals(instance1.particles.velocity, particles1evolved.velocity, 10)
        instance1.stop()

        instance2 = self.new_instance_of_an_optional_code(Rebound)
        instance2.parameters.epsilon_squared = 0.0 | nbody_system.length**2
        particles2c = particles2.copy()
        particles2c.subset = 0
        instance2.particles.add_particles(particles2c)
        instance2.evolve_model(0.1 | nbody_system.time)
        particles2evolved = particles_evolved[particles_evolved.subset == subset1]
        print("HHH:", particles_evolved.subset)
        print("p2e:", particles2evolved)
        print(instance2.particles)
        self.assertAlmostRelativeEquals(instance2.particles.position, particles2evolved.position, 10)
        self.assertAlmostRelativeEquals(instance2.particles.velocity, particles2evolved.velocity, 10)
        instance2.stop()

    def test9(self):
        instance = self.new_instance_of_an_optional_code(Rebound)
        # instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2

        particles = datamodel.Particles(10)
        particles.x = numpy.arange(0, 10, 1) | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [0, 1, 0] | nbody_system.speed
        particles[5].velocity = [0, 0, 0] | nbody_system.speed
        particles.radius = 0 | nbody_system.length
        particles[5].mass = 1 | nbody_system.mass

        instance.particles.add_particles(particles)
        self.assertEqual(instance.particles[5].mass, 1 | nbody_system.mass)
        self.assertEqual(instance.particles[6].mass, 0 | nbody_system.mass)

        instance.evolve_model(1 | nbody_system.time)
        self.assertEqual(instance.particles[5].mass, 1 | nbody_system.mass)
        self.assertEqual(instance.particles[6].mass, 0 | nbody_system.mass)
        self.assertEqual(instance.particles[5].x, 5 | nbody_system.length)
        self.assertAlmostRelativeEquals(instance.particles[6].x, 5.5403023 | nbody_system.length, 5)
        instance.particles.remove_particle(particles[4])
        self.assertEqual(instance.particles[4].mass, 1 | nbody_system.mass)
        self.assertEqual(instance.particles[5].mass, 0 | nbody_system.mass)
        instance.evolve_model(1 | nbody_system.time)
        self.assertEqual(instance.particles[4].x, 5 | nbody_system.length)
        instance.stop()

    def test10(self):
        instance = self.new_instance_of_an_optional_code(Rebound)
        # instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2

        particles = datamodel.Particles(1)
        particles.x = 0 | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [0, 2, 0] | nbody_system.speed
        particles.radius = 1 | nbody_system.length
        particles.mass = 1 | nbody_system.mass

        instance.particles.add_particles(particles)
        instance.evolve_model(4 | nbody_system.time)
        self.assertAlmostRelativeEquals(instance.particles[0].velocity, [0, 2, 0] | nbody_system.speed)
        self.assertAlmostRelativeEquals(instance.particles[0].position, [0, 2 * 4, 0] | nbody_system.length, 8)
        instance.stop()



"""
Runs several tests to determine how fast common operations
are in AMUSE.

to profile (in the amuse root directory):

./amuse.sh -m cProfile -s cumulative test/reports/speed_report.py --n_order==4 --test==speed_copy_to_set > profile.txt

"""


#import numpypy


from amuse.lab import *
#from amuse.datamodel import *
#from amuse.units import nbody_system
from amuse.support.thirdparty import texttable

import traceback

import subprocess
import os
import numpy
import time
import sys
import signal

from optparse import OptionParser

from mpi4py import MPI

from amuse.datamodel import ParticlesSuperset
class TimeoutException(Exception):
    pass
    
class SkipException(Exception):
    pass
    
class RunSpeedTests(object):
    
    @late
    def report_lines(self):
        return []
                
    @late
    def header_line(self):
        return ('action', 'duration\n(seconds)')
        
    @late
    def row_formatters(self):
        return (self.method_to_action, lambda x : x)
    
    @late
    def maximum_number_of_seconds_to_allow_per_test(self):
        return 1200
        
    def method_to_action(self, x):
        name = x.__name__
        name = name[len(self.method_prefix):]
        name = name.replace('_d_', '.')
        name = name.replace('_', ' ')
        return name
        
    def __init__(self, 
            total_number_of_points, 
            subset_number_of_points = -1, 
            name_of_the_method = None,
            include_code_tests = False,
            include_slow_tests = False,
            csv_output = False):
        self.total_number_of_points = total_number_of_points
        if subset_number_of_points < 0:
            subset_number_of_points = self.total_number_of_points
            
        self.subset_number_of_points = subset_number_of_points
        self.name_of_the_method = name_of_the_method
        self.include_code_tests = include_code_tests
        self.include_slow_tests = include_slow_tests
        self.csv_output = csv_output
    
    def handle_timeout(self, signum, frame):
        self.t1=-1
        self.t0=0
        raise TimeoutException("Test did not finish in allocated time frame")

        
    def start_measurement(self):
        self.t0 = time.time()
        signal.setitimer(signal.ITIMER_REAL, self.maximum_number_of_seconds_to_allow_per_test)
        signal.signal(signal.SIGALRM, self.handle_timeout)

        
    def end_measurement(self):
        self.t1 = time.time()
        signal.setitimer(signal.ITIMER_REAL, 0)
        
    def run(self):
        
        self.total_time = 0.0
    
        for x in self.names_of_testing_methods():
            if not self.name_of_the_method is None:
                if x != self.name_of_the_method:
                    continue
            method = getattr(self, x)
            print(self.row_formatters[0](method), '...', end=' ', file=sys.stderr)
            try:        
                method()
            except TimeoutException as ex:
                print("timed out,", ex, file=sys.stderr)
                continue
            except SkipException as ex:
                print("skipped,", ex, file=sys.stderr)
                continue
            except Exception as ex:
                print(ex)
                traceback.print_exc()
                self.t1=-1
                self.t0=0
                pass
            delta_time = self.t1-self.t0
            self.total_time = self.total_time + delta_time
            print(self.row_formatters[1](delta_time), file=sys.stderr)
            self.report_lines.append((method,delta_time))
        
        if self.csv_output:
            self.output_csv_line()
        else:
            self.output_table()
    
    def output_csv_line(self):
        line = []
        line.append('N')
        line.append('M')
        for x in self.report_lines:
            line.append(self.method_to_action(x[0]))
        print('#' + ','.join(line))
        
        line = []
        line.append(str(self.total_number_of_points))
        line.append(str(self.subset_number_of_points))
        for x in self.report_lines:
            line.append(str(x[1]))
        print(','.join(line))
        
    def output_table(self):
        lines = []
        
        lines.append(':run date:')
        lines.append('    {0}'.format(time.asctime()))
        lines.append('')
        lines.append(':number of points:')
        lines.append('    {0}'.format(self.total_number_of_points))
        lines.append('')
        
        for x in lines:
            print(x)
            
        table = texttable.Texttable()
        #table.set_deco(texttable.Texttable.HEADER)
        table.set_cols_dtype([
            't',  # text 
            'f',  # float (decimal)
        ])
        table.set_cols_align(["l", "r"])
        rows = []
        rows.append(self.header_line)
        rows.extend(self.report_lines_as_strings)
        table.add_rows(rows)
        print(table.draw())
        
            
    def names_of_testing_methods(self):
        for x in sorted(dir(type(self))):
            if x.startswith(self.method_prefix):
                yield x
    
    @late
    def method_prefix(self):
        return "speed_"
        
    @late
    def report_lines_as_strings(self):
        return list(self.iter_report_lines_as_strings())
    
    def iter_report_lines_as_strings(self):
        for line in self.report_lines:
            yield list(map(lambda x, formatter : formatter(x), line, self.row_formatters))
        
    def speed_make_plummer_sphere(self):
        """plummer sphere"""
        self.start_measurement()
        new_plummer_model(self.total_number_of_points)
        self.end_measurement()
        
    def speed_make_salpeter_mass_distribution(self):
        """plummer sphere"""
        self.start_measurement()
        new_salpeter_mass_distribution(self.total_number_of_points)
        self.end_measurement()
        
    def speed_make_salpeter_mass_distribution_nbody(self):
        """plummer sphere"""
        self.start_measurement()
        new_salpeter_mass_distribution_nbody(self.total_number_of_points)
        self.end_measurement()
        
    def speed_scale_plummer_sphere(self):
        input = new_plummer_model(self.total_number_of_points)
        
        self.is_slow_test()
        
        self.start_measurement()
        input.scale_to_standard()
        self.end_measurement()
        
    def is_slow_test(self):
        if not self.include_slow_tests:
            raise SkipException("slow tests disabled")
            
    def is_single_particle_test(self):
        if not self.include_slow_tests:
            raise SkipException("single particle tests disabled")
            
    def speed_calculate_potential_energy(self):
        
        self.is_slow_test()
        input = new_plummer_model(self.total_number_of_points)
        self.start_measurement()
        input.potential_energy(G=nbody_system.G)
        self.end_measurement()
        
        
    def speed_calculate_kinetic_energy(self):
        input = new_plummer_model(self.total_number_of_points)
        self.start_measurement()
        input.kinetic_energy()
        self.end_measurement()
        
    
    def speed_start_and_stop_BHTree_code(self):
        self.is_code_test()
        self.start_measurement()
        code = BHTree()
        code.stop()
        self.end_measurement()
    
    def speed_add_particles_to_code(self):
        self.is_code_test()
        
        code = BHTree()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        self.start_measurement()
        code.particles.add_particles(particles)
        self.end_measurement()
        code.stop()
        
    
    def speed_add_particles_to_code_SI(self):
        """plummer sphere"""
        self.is_code_test()
        
        converter = nbody.nbody_to_si(1 | units.parsec, self.total_number_of_points | units.MSun)
        code = BHTree(converter)
        particles = new_plummer_model(self.total_number_of_points, converter)
        particles.radius = 0| units.RSun
        self.start_measurement()
        code.particles.add_particles(particles)
        self.end_measurement()
        code.stop()
    
    
    def speed_copy_particles_from_code(self):
        self.is_code_test()
        code = BHTree()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        code.particles.add_particles(particles)
        channel = code.particles.new_channel_to(particles)
        self.start_measurement()
        channel.copy()
        self.end_measurement()
        code.stop()
    
    def speed_evolve_code_0_d_001_time_in_BHTree(self):
        """plummer sphere"""
    
        self.is_code_test()
        self.is_slow_test()
    
        code = BHTree()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        code.particles.add_particles(particles)
        self.start_measurement()
        code.evolve_model(0.001 | nbody.time)
        self.end_measurement()
        code.stop()
    
    def speed_evolve_code_0_d_001_time_in_Hermite(self):
        """plummer sphere"""
    
        self.is_code_test()
        self.is_slow_test()
    
        code = Hermite()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        code.particles.add_particles(particles)
        self.start_measurement()
        code.evolve_model(0.001 | nbody.time)
        self.end_measurement()
        code.stop()
       
    def is_code_test(self):
        if not self.include_code_tests:
            raise SkipException("code tests disabled") 

    @late
    def maximum_column_widths(self):
        maximums = [len(x) for x in self.header_line]
        for row in self.report_lines_as_strings:
            maximums = [max(len(x), y) for x,y in zip(row, maximums)]
        return maximums
    
    @late
    def column_widths(self):
        return [x + 2 for x in self.maximum_column_widths]
        
    def grid_table_row_separator_line(self, line_character = '-' ):
        parts = []
        for x in self.column_widths:
            parts.append('+')
            parts.append(line_character * x)
        parts.append('+')
        return ''.join(parts)
    
    
    def grid_table_row_line(self, row):
        parts = []
        for width, x in zip(self.column_widths, row):
            parts.append('|')
            parts.append(' ')
            parts.append(x.rjust(width-1))
        parts.append('|')
        return ''.join(parts)
    
    

    def speed_iterate_over_particles(self):
        
        self.is_single_particle_test()
        
        particles = Particles(self.total_number_of_points)
        particles.radius = 1.0 | nbody_system.length
        array = numpy.zeros(self.total_number_of_points, dtype=numpy.object)
        self.start_measurement()
        i = 0
        for x in particles:
            array[i] = x
            i += 1
        self.end_measurement()
    
    
    def speed_iterate_over_particles2(self):
        
        self.is_single_particle_test()
        
        class A(object):
            __slots__ = ('i', 'j', 'k' , 'l')
            __array_interface__ = {'shape':()}
            
            def __len__(self):
                raise AttributeError()
            def __iter__(self):
                raise AttributeError()
    
            
            def __init__(self, i, j = 10, k = 20, l = 24):
                self.i = i
                if i > 10:
                    self.j = j + 10
                else:
                    self.j = j
                
                self.k = k
                self.l = l
        
            def __getattr__(self, name_of_the_attribute):
                raise AttributeError("You tried to access attribute '{0}' but this attribute is not defined for this set.".format(name_of_the_attribute, ex))
    
                
        array = numpy.zeros(self.total_number_of_points, dtype=numpy.object)
        
        self.start_measurement()
        for x in range(self.total_number_of_points):
            array[x] = A(x,x,x,x)
        self.end_measurement()
    

    def speed_iterate_over_array(self):
        self.is_single_particle_test()
        
        class Test(object):
            def __init__(self):
                self.radius = 1.0
    
        particles = [Test() for x in range(self.total_number_of_points)]
        self.start_measurement()
        for x in particles:
            x.radius
        self.end_measurement()
    
    

    def speed_create_N_particles(self):
        self.is_single_particle_test()
        
        particles = Particles(self.total_number_of_points)
        particles.radius = 1.0 | nbody_system.length
        self.start_measurement()
        for x in range(self.total_number_of_points):
            Particle(x)
        self.end_measurement()
    
    def speed_copy_attributes_from_code(self):
        self.is_code_test()
        code = BHTree()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        code.particles.add_particles(particles)
        channel = code.particles.new_channel_to(particles)
        self.start_measurement()
        channel.copy()
        self.end_measurement()
        code.stop()
        
    def speed_copy_attributes_from_code_to_empty(self):
        self.is_code_test()
        code = BHTree()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        empty_particles = particles.empty_copy()
        
        code.particles.add_particles(particles)
        channel = code.particles.new_channel_to(empty_particles)
        self.start_measurement()
        channel.copy()
        self.end_measurement()
        code.stop()
        
    def speed_copy_mass_attribute_from_code_to_empty(self):
        self.is_code_test()
        code = BHTree()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        empty_particles = particles.empty_copy()
        
        code.particles.add_particles(particles)
        channel = code.particles.new_channel_to(empty_particles)
        self.start_measurement()
        channel.copy_attribute("mass","zmass")
        self.end_measurement()
        code.stop()
    
    def speed_copy_position_and_velocity_attributes_from_code_to_empty(self):
        self.is_code_test()
        code = BHTree()
        particles = new_plummer_model(self.total_number_of_points)
        particles.radius = 0| nbody.length
        empty_particles = particles.empty_copy()
        
        code.particles.add_particles(particles)
        channel1 = code.particles.new_channel_to(empty_particles)
        channel2 = empty_particles.new_channel_to(code.particles)
        self.start_measurement()
        channel1.copy_attributes(["x","y","z", "vx","vy","vz"])
        channel2.copy_attributes(["x","y","z"])
        self.end_measurement()
        code.stop()
        
    
    def speed_copy_to_superset(self):
        particles1 = new_plummer_model(self.total_number_of_points)
        particles2 = new_plummer_model(self.total_number_of_points)
        particles_all = ParticlesSuperset([particles1, particles2])
        
        empty_particles = particles_all.empty_copy()
        
        channel1 = particles_all.new_channel_to(empty_particles)
        channel2 = empty_particles.new_channel_to(particles_all)
        self.start_measurement()
        channel1.copy_attributes(["x","y","z"])
        channel2.copy_attributes(["x","y","z"])
        self.end_measurement()
    
    def speed_copy_to_set(self):
        particles_all = new_plummer_model(self.total_number_of_points * 2)
        empty_particles = particles_all.empty_copy()
        
        channel1 = particles_all.new_channel_to(empty_particles)
        channel2 = empty_particles.new_channel_to(particles_all)
        self.start_measurement()
        channel1.copy_attributes(["x","y","z"])
        channel2.copy_attributes(["x","y","z"])
        self.end_measurement()
        
    def speed_copy(self):
        particles_all = new_plummer_model(self.total_number_of_points)
        
        self.start_measurement()
        particles_all.copy_to_memory()
        self.end_measurement()
        
        
    def speed_copy_subset(self):
        particles_all = new_plummer_model(self.total_number_of_points)
        subset = particles_all[:self.subset_number_of_points]
        self.start_measurement()
        subset.copy_to_memory()
        self.end_measurement()
        
    
    def speed_select_array(self):
        particles_all = new_plummer_model(self.total_number_of_points)
        
        self.start_measurement()
        particles_selected = particles_all.select_array(lambda position: position.lengths() > 0.5 | nbody_system.length, ["position"])
        particles_selected.x
        self.end_measurement()
        
    def speed_select_with_get_item(self):
        particles_all = new_plummer_model(self.total_number_of_points)
        
        self.start_measurement()
        particles_selected = particles_all[particles_all.position.lengths() > 0.5 | nbody_system.length]
        self.end_measurement()

    def speed_iterate_over_quantity(self):
        
        lengths = numpy.arange(self.total_number_of_points) | nbody_system.length
        self.start_measurement()
        for x in range(self.total_number_of_points):
            lengths[x]
        self.end_measurement()
        
    def speed_add_particles(self):
        particles_to_add = new_plummer_model(self.total_number_of_points)
        step = self.total_number_of_points / 10
        particle_sets = []
        i = 0
        while i < self.total_number_of_points:
            j = i + step 
            if j > self.total_number_of_points:
                j = self.total_number_of_points
            particle_sets.append(particles_to_add[i:j].copy())
            i = j
            
        particles = Particles()
        self.start_measurement()
        for x in particle_sets:
            particles.add_particles(x)
        self.end_measurement()
        
def new_option_parser():
    result = OptionParser()
    result.add_option(
        "-n", 
        default = 2,
        dest="total_number_of_points",
        help="lenght of particle set",
        type="int"
    )
    result.add_option(
        "-m", 
        default = -1,
        dest="subset_number_of_points",
        help="lenght of particles in subset functions",
        type="int"
    )
    result.add_option(
        "--test", 
        default = None,
        dest="name_of_the_method",
        help="name of the test method to run",
        type="string"
    )
    result.add_option(
        "--code", 
        action="store_true",
        default=False,
        dest="include_code_tests",
        help="also run tests with codes"

    )
    result.add_option(
        "--slow", 
        action="store_true",
        default=False,
        dest="include_slow_tests",
        help="also run slow tests, scale N**2 or worse, for example calculating the potential energy"
    )
    result.add_option(
        "--csv", 
        action="store_true",
        default=False,
        help="display run as one comma separated line",
        dest="csv_output"

    )
    
    return result
    
    
    
if __name__ == '__main__':
    options, arguments = new_option_parser().parse_args()
    x = RunSpeedTests(**options.__dict__)
    x.run()











from amuse.test import amusetest
import numpy
import inspect
import collections


from amuse.rfi.tools import create_fortran
from amuse.rfi.core import *


class ForTestingInterface(CodeInterface):

    def __init__(self, exefile, **options):
        CodeInterface.__init__(self, exefile, **options)

    @legacy_function
    def echo_int():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in', dtype='int32', direction=function.IN)
        function.addParameter('int_out', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_double():
        function = LegacyFunctionSpecification()
        function.addParameter('double_in', dtype='float64', direction=function.IN)
        function.addParameter('double_out', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function


class TestGenerateAFortranStubStringFromASpecificationClass(amusetest.TestCase):

    def assertContainsString(self, string, substring):
        index = string.find(substring)
        if index < 0:
            self.fail("{0} not found in {1}".format(substring, string))

    def assertNotContainsString(self, string, substring):
        index = string.find(substring)
        if index >= 0:
            self.fail("{0} found in {1}".format(substring, string))

    def test1(self):
        x = create_fortran.GenerateAFortranStubStringFromASpecificationClass()
        x.specification_class = ForTestingInterface
        x.start()
        outputstring = x.result
        print(x.result)
        self.assertContainsString(outputstring, "function echo_int(int_in, int_out)")
        self.assertContainsString(outputstring, "integer :: echo_int")

    def test2(self):
        x = create_fortran.GenerateAFortranStubStringFromASpecificationClass()
        x.specification_class = ForTestingInterface
        x.start()
        outputstring = x.result
        self.assertNotContainsString(outputstring, "internal__")


class TestGenerateAFortranSourcecodeStringFromASpecificationClass(amusetest.TestCase):

    def assertContainsString(self, string, substring):
        index = string.find(substring)
        if index < 0:
            self.fail("{0} not found in {1}".format(substring, string))

    def assertNotContainsString(self, string, substring):
        index = string.find(substring)
        if index >= 0:
            self.fail("{0} found in {1}".format(substring, string))

    def test1(self):
        x = create_fortran.GenerateAFortranSourcecodeStringFromASpecificationClass()
        x.specification_class = ForTestingInterface
        x.start()
        outputstring = x.result
        self.assertContainsString(outputstring, "integer :: echo_int")
        self.assertContainsString(outputstring, "integer :: echo_double")
        self.assertContainsString(outputstring, "echo_int(")
        self.assertContainsString(outputstring, "echo_double(")

    def test2(self):
        x = create_fortran.GenerateAFortranSourcecodeStringFromASpecificationClass()
        x.specification_class = ForTestingInterface
        x.underscore_functions_from_specification_classes = [ForTestingInterface]
        x.start()
        outputstring = x.result
        self.assertContainsString(outputstring, "integer :: echo_int_")
        self.assertContainsString(outputstring, "integer :: echo_double_")
        self.assertContainsString(outputstring, "echo_int_(")
        self.assertContainsString(outputstring, "echo_double_(")



from amuse.test import amusetest
from amuse.units import units, nbody_system, constants

from amuse.ic.fractalcluster import new_fractal_cluster_model


class TestFractalCluster(amusetest.TestCase):

    def test1(self):
        print("First test: making a fractal cluster.")
        target_number_of_particles = 100
        parts = new_fractal_cluster_model(N=target_number_of_particles)
        self.assertEqual(len(parts), 100)

    def test2(self):
        print("test 2: test energy.")
        target_number_of_particles = 100
        parts = new_fractal_cluster_model(N=target_number_of_particles)
        ek = parts.kinetic_energy()
        ep = parts.potential_energy(G=nbody_system.G)
        self.assertAlmostEqual(ek/abs(ep), 0.5, 12)
        self.assertAlmostRelativeEqual(ep, -0.5 | nbody_system.energy, 12)

    def test3(self):
        print("test 3: test energy physical units.")
        target_number_of_particles = 100
        convert_nbody = nbody_system.nbody_to_si(1000 | units.MSun, 1 | units.parsec)
        parts = new_fractal_cluster_model(N=target_number_of_particles, convert_nbody=convert_nbody)
        ek = parts.kinetic_energy()
        ep = parts.potential_energy()
        self.assertAlmostEqual(ek/abs(ep), 0.5, 12)
        self.assertAlmostRelativeEqual(ep, -0.5 * constants.G * (1000 | units.MSun)**2 / (1.0 | units.parsec), 12)

    def test4(self):
        print("Test with masses")
        target_number_of_particles = 100
        masses = (range(1, 11) | units.MSun) * 1.0
        convert_nbody = nbody_system.nbody_to_si(1000 | units.MSun, 1 | units.parsec)
        particles = new_fractal_cluster_model(masses=masses, convert_nbody=convert_nbody, do_scale=True)

        ek = particles.kinetic_energy()
        ep = particles.potential_energy()
        self.assertEqual(len(particles), 10)
        self.assertAlmostEqual(particles.total_mass(), 1000 | units.MSun)  # Note: total_mass == converter's mass unit!
        self.assertAlmostEqual(masses.sum(), 55 | units.MSun)  # Note: total_mass != masses.sum()
        self.assertAlmostEqual(particles.center_of_mass(), [0, 0, 0] | units.parsec)
        self.assertAlmostEqual(particles.center_of_mass_velocity(), [0, 0, 0] | units.km / units.s)
        self.assertAlmostEqual(ek/ep, -0.5, 12)
        self.assertAlmostRelativeEqual(ek, (0.25 * constants.G * (1000 | units.MSun)**2 / (1.0 | units.parsec)).as_quantity_in(ek.unit), 12)

    def test5(self):
        print("Test with masses, with correct mass unit in converter")
        target_number_of_particles = 100
        masses = (range(1, 11) | units.MSun) * 1.0
        convert_nbody = nbody_system.nbody_to_si(masses.sum(), 1 | units.parsec)
        particles = new_fractal_cluster_model(masses=masses, convert_nbody=convert_nbody, do_scale=True)

        ek = particles.kinetic_energy()
        ep = particles.potential_energy()
        self.assertEqual(len(particles), 10)
        self.assertAlmostEqual(particles.total_mass(), 55 | units.MSun)  # Note: total_mass == converter's mass unit!
        self.assertAlmostEqual(particles.center_of_mass(), [0, 0, 0] | units.parsec)
        self.assertAlmostEqual(particles.center_of_mass_velocity(), [0, 0, 0] | units.km / units.s)
        self.assertAlmostEqual(ek/ep, -0.5, 12)
        self.assertAlmostRelativeEqual(ek, (0.25 * constants.G * (55 | units.MSun)**2 / (1.0 | units.parsec)).as_quantity_in(ek.unit), 12)

    def test6(self):
        print("Test fractal dimension.")
        number_of_particles = 1000
        for target_fractal_dimension in [1.6, 2.0, 2.5, 3.0]:
            particles = new_fractal_cluster_model(
                N=number_of_particles,
                fractal_dimension=target_fractal_dimension,
                do_scale=False, random_seed=1234321)
            self.assertAlmostRelativeEquals(particles.box_counting_dimension(),
                target_fractal_dimension, 1)
            self.assertAlmostRelativeEquals(particles.correlation_dimension(),
                target_fractal_dimension, 1)
